import uuid

from v2share import V2Data


def test_shadowsocks_link():
    ss = V2Data(
        "shadowsocks",
        "remark",
        "127.0.0.1",
        1234,
        password="1234",
    )
    assert ss.to_link() == "ss://Tm9uZToxMjM0QDEyNy4wLjAuMToxMjM0#remark"


def test_trojan_link():
    tj = V2Data("trojan", "remark", "127.0.0.1", 1234, password="1234")
    assert tj.to_link() == "trojan://1234@127.0.0.1:1234?security=none&type=tcp#remark"
    tj.tls = "tls"
    tj.alpn = "h2"
    tj.allow_insecure = True
    assert (
        tj.to_link()
        == "trojan://1234@127.0.0.1:1234?security=tls&type=tcp&allowInsecure=1&alpn=h2#remark"
    )


def test_vless_link():
    vl = V2Data(
        "vless",
        "remark",
        "127.0.0.1",
        1234,
        uuid=uuid.UUID("bb34fc3a-529d-473a-a3d9-1749b2116f2a"),
    )
    assert (
        vl.to_link()
        == "vless://bb34fc3a-529d-473a-a3d9-1749b2116f2a@127.0.0.1:1234?security=none&type=tcp#remark"
    )


def test_vmess_link():
    vm = V2Data(
        "vmess",
        "remark",
        "127.0.0.1",
        1234,
        uuid=uuid.UUID("bb34fc3a-529d-473a-a3d9-1749b2116f2a"),
    )
    assert (
        vm.to_link()
        == "vmess://eyJhZGQiOiAiMTI3LjAuMC4xIiwgImFpZCI6ICIwIiwgImlkIjogImJiMzRmYzNhLTUyOWQtNDczYS1hM2Q5LTE3NDliMjExNmYyYSIsICJuZXQiOiAidGNwIiwgInBvcnQiOiAxMjM0LCAicHMiOiAicmVtYXJrIiwgInNjeSI6ICJhdXRvIiwgInRscyI6ICJub25lIiwgInYiOiAiMiJ9"
    )
    tls_vm = V2Data(
        "vmess",
        "remark",
        "127.0.0.1",
        1234,
        uuid=uuid.UUID("bb34fc3a-529d-473a-a3d9-1749b2116f2a"),
        tls="tls",
        sni="something.com",
        alpn="h2",
    )
    assert (
        tls_vm.to_link()
        == "vmess://eyJhZGQiOiAiMTI3LjAuMC4xIiwgImFpZCI6ICIwIiwgImFscG4iOiAiaDIiLCAiaWQiOiAiYmIzNGZjM2EtNTI5ZC00NzNhLWEzZDktMTc0OWIyMTE2ZjJhIiwgIm5ldCI6ICJ0Y3AiLCAicG9ydCI6IDEyMzQsICJwcyI6ICJyZW1hcmsiLCAic2N5IjogImF1dG8iLCAic25pIjogInNvbWV0aGluZy5jb20iLCAidGxzIjogInRscyIsICJ2IjogIjIifQ=="
    )



import json
import uuid

from v2share import SingBoxConfig, V2Data


def test_sing_box_chaining():
    detour_config = V2Data(
        "vmess",
        "detour_outbound",
        "127.0.0.1",
        1234,
        uuid=uuid.UUID("bb34fc3a-529d-473a-a3d9-1749b2116f2a"),
    )
    main_config = V2Data(
        "vmess",
        "main_outbound",
        "127.0.0.1",
        1234,
        uuid=uuid.UUID("bb34fc3a-529d-473a-a3d9-1749b2116f2a"),
        next=detour_config,
    )
    sb = SingBoxConfig()
    sb.add_proxies([main_config])

    result = json.loads(sb.render())

    main_outbound, selector_outbounds = None, []
    for outbound in result["outbounds"]:
        if outbound["tag"] == "main_outbound":
            main_outbound = outbound

        if outbound["type"] == "selector" or outbound["type"] == "urltest":
            selector_outbounds.append(outbound)

    assert main_outbound["detour"] == "detour_outbound"
    for outbound in selector_outbounds:
        assert (
            "detour_outbound" not in outbound["outbounds"]
        )  # detour outbound should not be in the selectors



import base64
import json
import urllib.parse as urlparse
from dataclasses import dataclass, field
from typing import Optional, Dict, List
from uuid import UUID

from v2share._utils import filter_dict, set_path_early_data
from v2share.exceptions import ProtocolNotSupportedError


@dataclass
class XrayNoise:
    type: str
    packet: str
    delay: str


@dataclass
class XMuxSettings:
    max_concurrency: Optional[str] = None
    max_connections: Optional[str] = None
    max_reuse_times: Optional[str] = None
    max_lifetime: Optional[str] = None
    max_request_times: Optional[str] = None
    keep_alive_period: Optional[int] = None


@dataclass
class SplitHttpSettings:
    mode: Optional[str] = None
    no_grpc_header: Optional[bool] = None
    padding_bytes: Optional[str] = None
    xmux: Optional[XMuxSettings] = None


@dataclass
class SingBoxMuxSettings:
    max_connections: Optional[int] = None
    min_streams: Optional[int] = None
    max_streams: Optional[int] = None
    padding: Optional[bool] = None


@dataclass
class MuxCoolSettings:
    concurrency: Optional[int] = None
    xudp_concurrency: Optional[int] = None
    xudp_proxy_443: Optional[str] = None


@dataclass
class MuxSettings:
    protocol: str = "mux_cool"
    sing_box_mux_settings: Optional[SingBoxMuxSettings] = None
    mux_cool_settings: Optional[MuxCoolSettings] = None


@dataclass
class V2Data:
    protocol: str
    remark: str
    address: str
    port: int
    shadowsocks_method: Optional[str] = None
    uuid: Optional[UUID] = None
    vmess_security: str = "auto"
    password: Optional[str] = None
    ed25519: Optional[str] = None
    host: Optional[str] = None
    http_headers: Dict[str, str] = field(default_factory=dict)
    transport_type: str = "tcp"
    grpc_multi_mode: Optional[bool] = None
    grpc_user_agent: Optional[str] = None
    path: Optional[str] = None
    header_type: Optional[str] = None
    tls: str = "none"
    flow: Optional[str] = None
    sni: Optional[str] = None
    fingerprint: Optional[str] = None
    alpn: Optional[str] = None
    reality_pbk: Optional[str] = None
    reality_sid: Optional[str] = None
    reality_spx: Optional[str] = None
    client_address: Optional[str] = None
    fragment: bool = False
    fragment_packets: str = "tlshello"
    fragment_length: str = "100-200"
    fragment_interval: str = "10-20"
    xray_noises: Optional[List[XrayNoise]] = None
    early_data: Optional[int] = None
    mtu: Optional[int] = None
    dns_servers: Optional[List[str]] = None
    allowed_ips: Optional[List[str]] = None
    congestion_control: Optional[str] = None
    tuic_udp_relay_mode: Optional[str] = None
    shadowtls_version: Optional[int] = None
    disable_sni: bool = False
    allow_insecure: bool = False
    weight: int = 1
    next: Optional["V2Data"] = None
    splithttp_settings: Optional[SplitHttpSettings] = None
    mux_settings: Optional[MuxSettings] = None

    def _apply_tls_settings(self, payload):
        if self.tls in ["tls", "reality"]:
            payload.update(
                {
                    "sni": self.sni,
                    "fp": self.fingerprint,
                }
            )
            if self.protocol in ["vless", "trojan"]:
                payload.update({"allowInsecure": int(self.allow_insecure) or None})
        if self.tls == "tls":
            payload.update(
                {
                    "alpn": self.alpn,
                }
            )
        elif self.tls == "reality":
            payload.update(
                {
                    "pbk": self.reality_pbk,
                    "sid": self.reality_sid,
                    "spx": self.reality_spx,
                }
            )

    def _apply_trojan_vless_transport(self, payload):
        if self.transport_type == "grpc":
            transport_data = {
                "serviceName": self.path,
                "authority": self.host,
                "mode": "multi" if self.grpc_multi_mode else "gun",
            }
        elif self.transport_type == "kcp":
            transport_data = {"seed": self.path}
        elif self.transport_type == "quic":
            transport_data = {"key": self.path, "quicSecurity": self.host}
        else:
            transport_data = {"path": self.path, "host": self.host}
            if self.transport_type in {"ws", "httpupgrade"} and self.early_data:
                transport_data["path"] = set_path_early_data(
                    transport_data.get("path", "/"), self.early_data
                )
            if (
                self.transport_type == "splithttp"
                and self.splithttp_settings is not None
            ):
                transport_data["mode"] = self.splithttp_settings.mode
                transport_data["extra"] = filter_dict(
                    {
                        "headers": self.http_headers,
                        "xPaddingBytes": self.splithttp_settings.padding_bytes,
                        "noGRPCHeader": self.splithttp_settings.no_grpc_header,
                        "xmux": (
                            filter_dict(
                                {
                                    "maxConcurrency": self.splithttp_settings.xmux.max_concurrency,
                                    "maxConnections": self.splithttp_settings.xmux.max_connections,
                                    "cMaxReuseTimes": self.splithttp_settings.xmux.max_reuse_times,
                                    "cMaxLifetimeMs": self.splithttp_settings.xmux.max_lifetime,
                                    "hMaxRequestTimes": self.splithttp_settings.xmux.max_request_times,
                                    "hKeepAlivePeriod": self.splithttp_settings.xmux.keep_alive_period,
                                },
                                (None,),
                            )
                            if self.splithttp_settings.xmux is not None
                            else None
                        ),
                    },
                    (None,),
                )

        payload.update(transport_data)

    def to_link(self):
        if self.protocol == "shadowsocks":
            return (
                "ss://"
                + base64.b64encode(
                    (
                        f"{self.shadowsocks_method}:{self.password}"
                        + f"@{self.address}:{self.port}"
                    ).encode()
                ).decode()
                + f"#{urlparse.quote(self.remark)}"
            )

        if self.protocol == "vmess":
            payload = {
                "add": self.address,
                "aid": "0",
                "host": self.host,
                "id": str(self.uuid),
                "net": (
                    self.transport_type
                    if self.transport_type != "splithttp"
                    else "xhttp"
                ),
                "path": self.path,
                "port": self.port,
                "ps": self.remark,
                "scy": self.vmess_security,
                "tls": self.tls,
                "type": (
                    self.header_type
                    if self.transport_type != "splithttp"
                    or not self.splithttp_settings
                    or not self.splithttp_settings.mode
                    else self.splithttp_settings.mode
                ),
                "v": "2",
            }

            self._apply_tls_settings(payload)
            payload = filter_dict(payload, ("", None))
            return (
                "vmess://"
                + base64.b64encode(
                    json.dumps(payload, sort_keys=True).encode()
                ).decode()
            )

        if self.protocol in ["trojan", "vless"]:
            payload = {
                "security": self.tls,
                "type": (
                    self.transport_type
                    if self.transport_type != "splithttp"
                    else "xhttp"
                ),
                "headerType": self.header_type,
            }
            if self.protocol == "vless" and self.flow:
                payload.update({"flow": self.flow or ""})

            self._apply_trojan_vless_transport(payload)

            self._apply_tls_settings(payload)

            payload = filter_dict(payload, ("", None))

            passphrase = (
                self.uuid
                if self.protocol == "vless"
                else urlparse.quote(self.password, safe=":")
            )
            return (
                f"{self.protocol}://"
                + f"{passphrase}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{(urlparse.quote(self.remark))}"
            )

        if self.protocol == "hysteria2":
            payload = {"sni": self.sni}
            if self.header_type:
                payload.update({"obfs": self.header_type})
                payload.update({"obfs-password": self.path})
            if self.disable_sni:
                payload.update({"disable_sni": "1"})
            if self.allow_insecure:
                payload.update({"insecure": "1"})

            return (
                "hysteria2://"
                + f"{self.password}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{(urlparse.quote(self.remark))}"
            )

        if self.protocol == "wireguard":
            payload = {"publickey": self.path, "address": self.client_address}
            if isinstance(self.mtu, int):
                payload.update({"mtu": self.mtu})
            return (
                "wireguard://"
                + f"{self.ed25519}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{(urlparse.quote(self.remark))}"
            )

        if self.protocol == "tuic":
            payload = {
                "sni": self.sni,
                "udp_relay_mode": self.tuic_udp_relay_mode,
                "congestion_control": self.congestion_control,
                "alpn": self.alpn,
            }
            if self.allow_insecure:
                payload.update({"insecure": "1"})
            if self.disable_sni:
                payload.update({"disable_sni": "1"})

            payload = filter_dict(payload, ("", None))
            return (
                "tuic://"
                + f"{self.uuid}:{self.password}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{(urlparse.quote(self.remark))}"
            )
        raise ProtocolNotSupportedError



"""v2share utilities"""

from typing import Dict, Sequence, Any
from urllib.parse import parse_qs, urlsplit


def filter_dict(d: Dict[Any, Any], values: Sequence[Any]) -> dict:
    """
    removes pairs with null value from dictionary
    :param d: the dictionary
    :param values: values you want to be filtered
    :return: the filtered dictionary
    """
    return dict(filter(lambda p: p[1] not in values, d.items()))


def set_path_early_data(path: str, early_data: int):
    parsed = urlsplit(path)
    query_params = parse_qs(parsed.query)

    query_params["ed"] = [str(early_data)]

    query_string = "&".join(
        "&".join([f"{key}={v}" for v in values]) for key, values in query_params.items()
    )

    return parsed.path + "?" + query_string



from .clash import ClashConfig
from .clashmeta import ClashMetaConfig
from .data import V2Data
from .singbox import SingBoxConfig
from .wireguard import WireGuardConfig
from .xray import XrayConfig

__all__ = [
    "V2Data",
    "SingBoxConfig",
    "ClashConfig",
    "ClashMetaConfig",
    "XrayConfig",
    "WireGuardConfig",
]



import json
import random
from importlib import resources
from typing import List

from v2share._utils import filter_dict
from v2share.base import BaseConfig
from v2share.data import V2Data
from v2share.exceptions import ProtocolNotSupportedError, TransportNotSupportedError


class SingBoxConfig(BaseConfig):
    chaining_support = True
    supported_protocols = [
        "shadowsocks",
        "vmess",
        "trojan",
        "vless",
        "hysteria2",
        "wireguard",
        "shadowtls",
        "tuic",
    ]
    supported_transports = [
        "tcp",
        "ws",
        "quic",
        "httpupgrade",
        "grpc",
        "http",
        "splithttp",
        None,
    ]

    def __init__(self, template_path: str = None, swallow_errors=True):
        if not template_path:
            template_path = resources.files("v2share.templates") / "singbox.json"
        with open(template_path) as f:
            self._template_data = f.read()
        self._outbounds = []
        self._swallow_errors = swallow_errors
        self._configs: List[V2Data] = []

    def render(self, sort: bool = True, shuffle: bool = False):
        if shuffle is True:
            configs = random.sample(self._configs, len(self._configs))
        elif sort is True:
            configs = sorted(self._configs, key=lambda c: c.weight)
        else:
            configs = self._configs

        result = json.loads(self._template_data)

        blackset = set()
        for config in configs:
            c = config
            while True:
                outbound = self.create_outbound(c)
                if c.next:
                    outbound["detour"] = config.next.remark
                    blackset.add(c.next.remark)
                    c = config.next
                    result["outbounds"].append(outbound)
                else:
                    result["outbounds"].append(outbound)
                    break

        urltest_types = [
            "hysteria2",
            "vmess",
            "vless",
            "trojan",
            "shadowsocks",
            "wireguard",
            "tuic",
            "shadowtls",
        ]
        urltest_tags = [
            outbound["tag"]
            for outbound in result["outbounds"]
            if outbound["type"] in urltest_types and outbound["tag"] not in blackset
        ]
        selector_types = [
            "hysteria2",
            "vmess",
            "vless",
            "trojan",
            "shadowsocks",
            "wireguard",
            "tuic",
            "shadowtls",
            "urltest",
        ]
        selector_tags = [
            outbound["tag"]
            for outbound in result["outbounds"]
            if outbound["type"] in selector_types and outbound["tag"] not in blackset
        ]

        for outbound in result["outbounds"]:
            if outbound.get("type") == "urltest":
                if not outbound.get("outbounds"):
                    outbound["outbounds"] = urltest_tags

        for outbound in result["outbounds"]:
            if outbound.get("type") == "selector":
                if not outbound.get("outbounds"):
                    outbound["outbounds"] = selector_tags

        return json.dumps(result, indent=4)

    @staticmethod
    def tls_config(
        sni=None, fp=None, tls=None, pbk=None, sid=None, alpn=None, ais=None
    ):
        config = {}
        if tls in ["tls", "reality"]:
            config["enabled"] = True

        if sni is not None:
            config["server_name"] = sni

        if tls == "tls" and ais:
            config["insecure"] = ais

        if tls == "reality":
            config["reality"] = {"enabled": True}
            if pbk:
                config["reality"]["public_key"] = pbk
            if sid:
                config["reality"]["short_id"] = sid

        if fp:
            config["utls"] = {"enabled": True, "fingerprint": fp}

        if alpn:
            config["alpn"] = [alpn] if not isinstance(alpn, list) else alpn

        return config

    @staticmethod
    def transport_config(
        transport_type="tcp",
        host=None,
        path=None,
        http_method=None,
        headers=None,
        early_data=None,
    ):
        if headers is None:
            headers = {}

        transport_config = {"type": transport_type}

        if transport_type in {"http", "tcp", "splithttp"}:
            transport_config["type"] = "http"
            transport_config["headers"] = headers
            if host:
                transport_config["host"] = [host]
            if path:
                transport_config["path"] = path
            if http_method:
                transport_config["method"] = http_method
        elif transport_type == "ws":
            transport_config["headers"] = headers
            if path:
                if early_data:
                    transport_config["early_data_header_name"] = (
                        "Sec-WebSocket-Protocol"
                    )
                    transport_config["max_early_data"] = early_data
                transport_config["path"] = path
            if host:
                transport_config["headers"]["Host"] = host
        elif transport_type == "httpupgrade":
            transport_config["headers"] = headers
            if host:
                transport_config["host"] = host
            if path:
                transport_config["path"] = path
        elif transport_type == "grpc":
            if path:
                transport_config["service_name"] = path

        return transport_config

    @staticmethod
    def create_outbound(config: V2Data):
        outbound = {
            "type": config.protocol,
            "tag": config.remark,
            "server": config.address,
            "server_port": config.port,
        }
        if (
            config.protocol == "vless"
            and config.flow
            and config.tls in ["tls", "reality"]
        ):
            outbound["flow"] = config.flow

        if config.transport_type in [
            "ws",
            "quic",
            "grpc",
            "httpupgrade",
            "http",
            "splithttp",
        ] or (config.transport_type == "tcp" and config.header_type == "http"):
            outbound["transport"] = SingBoxConfig.transport_config(
                transport_type=config.transport_type,
                host=config.host,
                path=config.path,
                headers=config.http_headers,
                early_data=config.early_data,
            )

        if config.tls in ("tls", "reality"):
            outbound["tls"] = SingBoxConfig.tls_config(
                sni=config.sni,
                fp=config.fingerprint,
                tls=config.tls,
                pbk=config.reality_pbk,
                sid=config.reality_sid,
                alpn=config.alpn,
                ais=config.allow_insecure,
            )

        if config.protocol in ["vless", "vmess"]:
            outbound["uuid"] = str(config.uuid)

        elif config.protocol == "trojan":
            outbound["password"] = config.password

        elif config.protocol == "shadowsocks":
            outbound["password"] = config.password
            outbound["method"] = config.shadowsocks_method
        elif config.protocol == "hysteria2":
            outbound["password"] = config.password
            if config.header_type:
                outbound["obfs"] = {"type": config.header_type, "password": config.path}
        elif config.protocol == "wireguard":
            outbound.update(
                {
                    "private_key": config.ed25519,
                    "local_address": [config.client_address],
                    "mtu": config.mtu,
                    "peers": [
                        {
                            "server": config.address,
                            "port": config.port,
                            "public_key": config.path,
                            "allowed_ips": config.allowed_ips or ["0.0.0.0/0", "::/0"],
                        }
                    ],
                }
            )
        elif config.protocol == "shadowtls":
            if config.shadowtls_version:
                outbound["version"] = config.shadowtls_version
                if config.shadowtls_version in {2, 3}:
                    outbound["password"] = config.password
        elif config.protocol == "tuic":
            outbound["password"] = config.password
            outbound["uuid"] = str(config.uuid)

        if config.mux_settings is not None and config.mux_settings.protocol in {
            "h2mux",
            "yamux",
            "smux",
        }:
            outbound["multiplex"] = {"enabled": True, "protocol": config.mux_settings.protocol}
            if config.mux_settings.sing_box_mux_settings is not None:
                additional_mux_settings = filter_dict(
                    {
                        "max_connections": config.mux_settings.sing_box_mux_settings.max_connections,
                        "min_streams": config.mux_settings.sing_box_mux_settings.min_streams,
                        "max_streams": config.mux_settings.sing_box_mux_settings.max_streams,
                        "padding": config.mux_settings.sing_box_mux_settings.padding,
                    },
                    (None,),
                )
                outbound["multiplex"].update(additional_mux_settings)
        return outbound

    def add_proxies(self, proxies: List[V2Data]):
        for proxy in proxies:
            # validation
            if (
                unsupported_transport := proxy.transport_type
                not in self.supported_transports
            ) or (
                unsupported_protocol := proxy.protocol not in self.supported_protocols
            ):
                if self._swallow_errors:
                    continue
                if unsupported_transport:
                    raise TransportNotSupportedError
                if unsupported_protocol:
                    raise ProtocolNotSupportedError

            self._configs.append(proxy)



import json
import random
from dataclasses import asdict
from importlib import resources
from typing import List, Optional

from v2share._utils import set_path_early_data, filter_dict
from v2share.base import BaseConfig
from v2share.data import V2Data, XrayNoise, SplitHttpSettings
from v2share.exceptions import ProtocolNotSupportedError, TransportNotSupportedError


class XrayConfig(BaseConfig):
    chaining_support = True
    supported_transports = [
        "tcp",
        "kcp",
        "ws",
        "http",
        "quic",
        "grpc",
        "httpupgrade",
        "splithttp",
        None,
    ]
    supported_protocols = ["vmess", "vless", "trojan", "shadowsocks", "wireguard"]

    def __init__(
        self,
        template_path: str = None,
        swallow_errors=True,
    ):
        self.config = []
        self._configs = []
        self._swallow_errors = swallow_errors
        if not template_path:
            template_path = resources.files("v2share.templates") / "xray.json"
        with open(template_path) as f:
            self._template = f.read()

    def add_proxies(self, proxies: List[V2Data]):
        for proxy in proxies:
            # validation
            if (
                unsupported_transport := proxy.transport_type
                not in self.supported_transports
            ) or (
                unsupported_protocol := proxy.protocol not in self.supported_protocols
            ):
                if self._swallow_errors:
                    continue
                if unsupported_transport:
                    raise TransportNotSupportedError
                if unsupported_protocol:
                    raise ProtocolNotSupportedError

            self._configs.append(proxy)

    def render(self, sort: bool = True, shuffle: bool = False):
        configs: List[V2Data] = []
        if shuffle is True:
            configs = random.sample(self._configs, len(self._configs))
        elif sort is True:
            configs = sorted(self._configs, key=lambda config: config.weight)
        else:
            configs = self._configs

        xray_configs = []
        for data in configs:
            outbounds = self.create_outbounds(data)
            json_template = json.loads(self._template)
            complete_config = {
                **json_template,
                **{
                    "remarks": data.remark,
                    "outbounds": outbounds + json_template["outbounds"],
                },
            }
            xray_configs.append(complete_config)
        return json.dumps(xray_configs, indent=4)

    @staticmethod
    def create_outbounds(data: V2Data):
        outbound = {"tag": data.remark, "protocol": data.protocol}
        outbounds = [outbound]
        dialer_proxy = None

        if data.protocol == "vmess":
            outbound["settings"] = XrayConfig.vmess_config(
                address=data.address, port=data.port, uuid=str(data.uuid)
            )
        elif data.protocol == "vless":
            if data.tls in {"reality", "tls"}:
                flow = data.flow or ""
            else:
                flow = ""
            outbound["settings"] = XrayConfig.vless_config(
                address=data.address, port=data.port, uuid=str(data.uuid), flow=flow
            )

        elif data.protocol == "trojan":
            outbound["settings"] = XrayConfig.trojan_config(
                address=data.address, port=data.port, password=data.password
            )

        elif data.protocol == "shadowsocks":
            outbound["settings"] = XrayConfig.shadowsocks_config(
                address=data.address,
                port=data.port,
                password=data.password,
                method=data.shadowsocks_method,
            )
        elif data.protocol == "wireguard":
            outbound["settings"] = XrayConfig.wireguard_config(
                address=data.address,
                port=data.port,
                peer_public_key=data.path,
                private_key=data.ed25519,
                mtu=data.mtu,
                allowed_ips=data.allowed_ips or ["0.0.0.0/0", "::/0"],
                keepalive=25,
                client_address=[data.client_address],
            )

        if data.protocol != "wireguard":
            if data.next:
                next_outbounds = XrayConfig.create_outbounds(data.next)
                outbounds.extend(next_outbounds)
                dialer_proxy = next_outbounds[0]["tag"]
            elif data.fragment:
                fragment_outbound = XrayConfig.make_fragment_outbound(
                    data.fragment_packets, data.fragment_length, data.fragment_interval
                )
                outbounds.append(fragment_outbound)
                dialer_proxy = fragment_outbound["tag"]
            elif data.xray_noises:
                noisy_outbound = XrayConfig.make_noisy_outbound(data.xray_noises)
                outbounds.append(noisy_outbound)
                dialer_proxy = noisy_outbound["tag"]

            outbound["streamSettings"] = XrayConfig.make_stream_settings(
                net=data.transport_type,
                tls=data.tls,
                sni=data.sni,
                host=data.host,
                path=data.path,
                alpn=data.alpn,
                fp=data.fingerprint,
                pbk=data.reality_pbk,
                sid=data.reality_sid,
                ais=data.allow_insecure,
                header_type=data.header_type,
                grpc_multi_mode=data.grpc_multi_mode,
                dialer_proxy=dialer_proxy,
                headers=data.http_headers,
                early_data=data.early_data,
                splithttp_settings=data.splithttp_settings,
            )

        if data.mux_settings is not None and data.mux_settings.protocol == "mux_cool":
            mux_config = {"enabled": True}
            if data.mux_settings.mux_cool_settings is not None:
                mux_config.update(
                    filter_dict(
                        {
                            "concurrency": data.mux_settings.mux_cool_settings.concurrency,
                            "xudpConcurrency": data.mux_settings.mux_cool_settings.xudp_concurrency,
                            "xudpProxyUDP443": data.mux_settings.mux_cool_settings.xudp_proxy_443,
                        },
                        (None,),
                    )
                )

            outbound["mux"] = mux_config

        return outbounds

    @staticmethod
    def tls_config(sni, fingerprint, alpn=None, ais=False):
        tls_settings = {"serverName": sni, "allowInsecure": ais or False}
        if alpn is None:
            alpn = ["h2", "http/1.1"]

        tls_settings["alpn"] = alpn

        if fingerprint:
            tls_settings["fingerprint"] = fingerprint

        return tls_settings

    @staticmethod
    def reality_config(public_key, short_id, sni, fingerprint="", spiderx=""):
        return {
            "serverName": sni,
            "fingerprint": fingerprint,
            "show": False,
            "publicKey": public_key,
            "shortId": short_id,
            "spiderX": spiderx,
        }

    @staticmethod
    def ws_config(path=None, host=None, headers=None, early_data=None):
        if headers is None:
            headers = {}

        ws_settings = {"headers": headers}
        if path:
            ws_settings["path"] = path

        if early_data:
            ws_settings["path"] = set_path_early_data(
                ws_settings.get("path", "/"), early_data
            )

        if host:
            ws_settings["host"] = host

        return ws_settings

    @staticmethod
    def httpupgrade_config(path=None, host=None, headers=None, early_data=None):
        if headers is None:
            headers = {}

        httpupgrade_settings = {"headers": headers}
        if path:
            httpupgrade_settings["path"] = path
        if early_data:
            httpupgrade_settings["path"] = set_path_early_data(
                httpupgrade_settings.get("path", "/"), early_data
            )
        if host:
            httpupgrade_settings["host"] = host

        return httpupgrade_settings

    @staticmethod
    def grpc_config(
        authority=None, service_name=None, multi_mode=None, user_agent=None
    ):
        grpc_settings = {}
        if multi_mode is not None:
            grpc_settings["multiMode"] = multi_mode
        if service_name is not None:
            grpc_settings["serviceName"] = service_name
        if authority is not None:
            grpc_settings["authority"] = authority
        if user_agent is not None:
            grpc_settings["user_agent"] = user_agent
        return grpc_settings

    @staticmethod
    def tcp_http_config(header_type=None, host=None, path=None):
        if header_type is None:
            header_type = "none"
        if path is None:
            path = "/"

        tcp_settings = {
            "header": {
                "type": header_type,
            }
        }

        if header_type != "none":
            tcp_settings["header"]["request"] = {
                "headers": {
                    "Host": [host] if host else [],
                },
                "path": [path],
            }

        return tcp_settings

    @staticmethod
    def h2_config(path=None, host=None, headers=None):
        if host is None:
            host = []
        if path is None:
            path = "/"
        if headers is None:
            headers = {}

        http_settings = {"path": path, "host": host}
        if headers:
            http_settings["headers"] = headers

        return http_settings

    @staticmethod
    def splithttp_config(
        path=None, host=None, headers=None, settings: Optional[SplitHttpSettings] = None
    ):
        if host is None:
            host = []
        if path is None:
            path = "/"
        if headers is None:
            headers = {}

        splithttp_settings = {"path": path, "headers": headers}

        if host:
            splithttp_settings["host"] = host

        if settings is not None:
            if settings.mode is not None:
                splithttp_settings["mode"] = settings.mode
            if settings.no_grpc_header is not None:
                splithttp_settings["noGRPCHeader"] = settings.no_grpc_header
            if settings.xmux is not None:
                splithttp_settings["xmux"] = filter_dict(
                    {
                        "maxConcurrency": settings.xmux.max_concurrency,
                        "maxConnections": settings.xmux.max_connections,
                        "cMaxReuseTimes": settings.xmux.max_reuse_times,
                        "cMaxLifetimeMs": settings.xmux.max_lifetime,
                        "hMaxRequestTimes": settings.xmux.max_request_times,
                        "hKeepAlivePeriod": settings.xmux.keep_alive_period,
                    },
                    (None,),
                )
        return splithttp_settings

    @staticmethod
    def quic_config(security="none", key=None, header_type="none"):
        quic_settings = {
            "security": security,
            "header": {"type": header_type},
        }

        if key:
            quic_settings["key"] = key

        return quic_settings

    @staticmethod
    def kcp_config(seed=None, header_type=None) -> dict:
        kcp_settings = {
            "header": {},
        }

        if seed:
            kcp_settings["seed"] = seed
        if header_type:
            kcp_settings["header"]["type"] = header_type

        return kcp_settings

    @staticmethod
    def make_stream_settings(
        net="tcp",
        path=None,
        host=None,
        tls="none",
        sni="",
        fp=None,
        alpn="h2,http/1.1",
        pbk=None,
        sid=None,
        ais=False,
        header_type=None,
        grpc_multi_mode=False,
        grpc_user_agent=None,
        dialer_proxy=None,
        headers=None,
        early_data=None,
        splithttp_settings=None,
    ) -> dict:
        if headers is None:
            headers = {}

        stream_settings = {"network": net}

        if net:
            if net in {"ws", "websocket"}:
                stream_settings["wsSettings"] = XrayConfig.ws_config(
                    path=path, host=host, headers=headers, early_data=early_data
                )
            elif net in {"grpc", "gun"}:
                stream_settings["grpcSettings"] = XrayConfig.grpc_config(
                    authority=host,
                    service_name=path,
                    multi_mode=grpc_multi_mode,
                    user_agent=grpc_user_agent,
                )
            elif net in {"h2", "http"}:
                stream_settings["httpSettings"] = XrayConfig.h2_config(
                    path=path, host=[host], headers=headers
                )
            elif net in {"kcp", "mkcp"}:
                stream_settings["kcpSettings"] = XrayConfig.kcp_config(
                    seed=path, header_type=header_type
                )
            elif net == "tcp":
                stream_settings["tcpSettings"] = XrayConfig.tcp_http_config(
                    header_type=header_type, host=host, path=path
                )
            elif net == "quic":
                stream_settings["quicSettings"] = XrayConfig.quic_config(
                    security=host, key=path, header_type=header_type
                )
            elif net == "httpupgrade":
                stream_settings["httpupgradeSettings"] = XrayConfig.httpupgrade_config(
                    path=path, host=host, headers=headers, early_data=early_data
                )
            elif net == "splithttp":
                stream_settings["splithttpSettings"] = XrayConfig.splithttp_config(
                    path=path,
                    host=host,
                    headers=headers,
                    settings=splithttp_settings,
                )

            if tls == "tls":
                stream_settings["security"] = "tls"
                stream_settings["tlsSettings"] = XrayConfig.tls_config(
                    sni=sni,
                    fingerprint=fp,
                    alpn=alpn.split(",") if alpn else None,
                    ais=ais,
                )
            elif tls == "reality":
                stream_settings["security"] = "reality"
                stream_settings["realitySettings"] = XrayConfig.reality_config(
                    pbk, sid, sni, fingerprint=fp
                )
            else:
                stream_settings["security"] = "none"

        if dialer_proxy:
            stream_settings["sockopt"] = {"dialerProxy": dialer_proxy}

        return stream_settings

    @staticmethod
    def vmess_config(address: str, port: int, uuid: str) -> dict:
        return {
            "vnext": [
                {
                    "address": address,
                    "port": port,
                    "users": [
                        {
                            "id": uuid,
                            "security": "auto",
                        }
                    ],
                }
            ]
        }

    @staticmethod
    def vless_config(address, port, uuid, flow="") -> dict:
        return {
            "vnext": [
                {
                    "address": address,
                    "port": port,
                    "users": [
                        {
                            "id": uuid,
                            "encryption": "none",
                            "flow": flow,
                        }
                    ],
                }
            ]
        }

    @staticmethod
    def trojan_config(address, port, password) -> dict:
        return {
            "servers": [
                {
                    "address": address,
                    "port": port,
                    "password": password,
                    "email": "kiomars",
                }
            ]
        }

    @staticmethod
    def shadowsocks_config(address: str, port: int, password: str, method: str) -> dict:
        return {
            "servers": [
                {
                    "address": address,
                    "port": port,
                    "password": password,
                    "email": "kiomars",
                    "method": method,
                    "uot": False,
                }
            ]
        }

    @classmethod
    def wireguard_config(
        cls,
        address: str,
        port: int,
        peer_public_key: str,
        private_key: str,
        client_address: List,
        mtu: int,
        allowed_ips: List[str],
        keepalive: int,
    ) -> dict:
        return {
            "secretKey": private_key,
            "address": client_address,
            "peers": [
                {
                    "endpoint": address + ":" + str(port),
                    "publicKey": peer_public_key,
                    "allowedIPs": allowed_ips,
                    "keepAlive": keepalive,
                }
            ],
            "mtu": mtu,
        }

    @staticmethod
    def make_fragment_outbound(packets, length, interval):
        return {
            "tag": "fragment_out",
            "protocol": "freedom",
            "settings": {
                "fragment": {"packets": packets, "length": length, "interval": interval}
            },
        }

    @staticmethod
    def make_noisy_outbound(xray_noises: List[XrayNoise]):
        return {
            "tag": "noisy_out",
            "protocol": "freedom",
            "settings": {"noises": [asdict(noise) for noise in xray_noises]},
        }



import json
from typing import List

from v2share import V2Data
from v2share.base import BaseConfig
from v2share.exceptions import TransportNotSupportedError, ProtocolNotSupportedError

supported_protocols = ["wireguard"]
supported_transports = [None]


class WireGuardConfig(BaseConfig):
    def __init__(self, swallow_errors=True):
        self._swallow_errors = swallow_errors
        self._configs: List[V2Data] = []

    def render(self, sort: bool, shuffle: bool) -> str:
        result = []
        for proxy in self._configs:
            config = "[Interface]\nAddress = " + proxy.client_address + "\n"
            config += "PrivateKey = " + proxy.ed25519 + "\n"
            if proxy.mtu is not None:
                config += "MTU = " + str(proxy.mtu) + "\n"
            if proxy.dns_servers:
                config += "DNS = " + ",".join(proxy.dns_servers) + "\n"
            config += "\n"
            config += "[Peer]\nPublicKey = " + proxy.path + "\n"
            config += "Endpoint = " + proxy.address + ":" + str(proxy.port) + "\n"
            config += "AllowedIPs = "
            if not proxy.allowed_ips:
                config += "0.0.0.0/0, ::/0" + "\n"
            else:
                config += ", ".join(proxy.allowed_ips) + "\n"
            config += "PersistentKeepalive = 25\n"
            result.append({"remark": proxy.remark, "config": config})
        return json.dumps(result)

    def add_proxies(self, proxies: List[V2Data]) -> None:
        for proxy in proxies:
            # validation
            if (
                unsupported_transport := proxy.transport_type
                not in supported_transports
            ) or (unsupported_protocol := proxy.protocol not in supported_protocols):
                if self._swallow_errors:
                    continue
                if unsupported_transport:
                    raise TransportNotSupportedError
                if unsupported_protocol:
                    raise ProtocolNotSupportedError

            self._configs.append(proxy)



import json
import uuid

from v2share import XrayConfig, V2Data


def test_xray_chaining():
    detour_config = V2Data(
        "vmess",
        "detour_outbound",
        "127.0.0.1",
        1234,
        uuid=uuid.UUID("bb34fc3a-529d-473a-a3d9-1749b2116f2a"),
    )
    main_config = V2Data(
        "vmess",
        "main_outbound",
        "127.0.0.1",
        1234,
        uuid=uuid.UUID("bb34fc3a-529d-473a-a3d9-1749b2116f2a"),
        next=detour_config,
    )
    x = XrayConfig()
    x.add_proxies([main_config])
    result = json.loads(x.render())

    main_outbound = None
    for outbound in result[0]["outbounds"]:
        if outbound["tag"] == "main_outbound":
            main_outbound = outbound
            break
    assert (
        main_outbound["streamSettings"]["sockopt"]["dialerProxy"] == "detour_outbound"
    )



class V2ShareError(Exception):
    """Base class for all v2share exceptions"""


class NotSupportedError(V2ShareError):
    """indicating that the proxy is somehow not supported by v2share"""


class TransportNotSupportedError(NotSupportedError):
    """indicating that the specified transport is not supported"""


class ProtocolNotSupportedError(NotSupportedError):
    """indicating that the protocol is not supported"""



import random
from typing import List

from v2share.base import BaseConfig
from v2share.data import V2Data
from v2share.exceptions import TransportNotSupportedError, ProtocolNotSupportedError


class LinksConfig(BaseConfig):
    supported_transports = [
        "tcp",
        "kcp",
        "ws",
        "http",
        "quic",
        "grpc",
        "httpupgrade",
        "splithttp",
        None,
    ]
    supported_protocols = [
        "vmess",
        "vless",
        "trojan",
        "shadowsocks",
        "hysteria2",
        "wireguard",
        "tuic",
    ]

    def __init__(self, swallow_errors=True):
        self._configs: List[V2Data] = []
        self._swallow_errors = swallow_errors

    def render(self, sort: bool = True, shuffle: bool = False):
        if shuffle is True:
            configs = random.sample(self._configs, len(self._configs))
        elif sort is True:
            configs = sorted(self._configs, key=lambda config: config.weight)
        else:
            configs = self._configs

        links = [config.to_link() for config in configs]
        return "\n".join(links)

    def add_proxies(self, proxies: List[V2Data]):
        for proxy in proxies:
            # validation
            if (
                unsupported_transport := proxy.transport_type
                not in self.supported_transports
            ) or (
                unsupported_protocol := proxy.protocol not in self.supported_protocols
            ):
                if self._swallow_errors:
                    continue
                if unsupported_transport:
                    raise TransportNotSupportedError
                if unsupported_protocol:
                    raise ProtocolNotSupportedError

            self._configs.append(proxy)
